package assignment2;

import java.util.Arrays;
import java.util.NoSuchElementException;

public class MyHashTable<T> implements A2HashTable<T> {
	private Container<T>[] table;
	private int size = 0;
	private int numberInserted = 0;
	final private double loadFactor;
	
	private class Container<U> {
		public Container(U value) {
			this.value = value;
		}
		private U value;
		private boolean isDeleted = false;
	}
	
	public MyHashTable(double loadFactor) {
		if (loadFactor <= 0 || loadFactor > 0.5) throw new IllegalArgumentException("Illegal load factor. 0 < lf <= 0.5 please.");
		table = (Container<T>[]) (new Container[getNextPrime()]);
		this.loadFactor = loadFactor;
	}

	@Override
	public void insert(T element) {
		int pos = findPos(element);
		
		if (table[pos] != null && table[pos].value.equals(element)) {
			if (table[pos].isDeleted == true) {
				table[pos].isDeleted = false;
				size+=1;
			}
			return;
		}
		
		
		table[pos] = new Container<T>(element);
		size += 1;
		numberInserted += 1;
		if ((numberInserted / (double)table.length) >= loadFactor) {
			rehash();
		}
	}
	
	private void rehash() {
		Container<T>[] oldTable = table;
		table = (Container<T>[]) (new Container[getNextPrime()]);
		numberInserted = 0;
		size = 0;
		for(int i = 0; i < oldTable.length; i++) {
			if (oldTable[i] != null && oldTable[i].isDeleted == false) insert(oldTable[i].value);
		}
	}
	
	private int findPos(T element) {
		int pos = Math.abs(element.hashCode())%table.length;
		int offset = 1;
		while( !(table[pos] == null) ) {
			if (table[pos].value.equals(element)) return pos;
			pos += offset;	//equivalent to pos = startPos + loops*loops  --- this is quadratic
			offset += 2;	//idea from http://users.cis.fiu.edu/~weiss/dsaajava3/code/QuadraticProbingHashTable.java
			while(pos >= table.length) pos -= table.length;
		}
		return pos;
	}

	@Override
	public void delete(T element) {
		int pos = findPos(element);	
		if (table[pos] == null || table[pos].isDeleted == true) {
			throw new NoSuchElementException();
		} 
		
		table[pos].isDeleted = true;
		size--;
	}

	@Override
	public boolean contains(T element) {
		int pos = findPos(element);
		return ((table[pos] != null) && (table[pos].isDeleted == false));
	}
	
	public int size() {
		return size;
	}
	public int numberInserted() {
		return numberInserted;
	}
	public double loadFactor() {
		return numberInserted / (double)table.length;
	}
	public void printTable() {
		Arrays.stream(table).map((Container c) -> c == null ? "|N" : "|"+ c.value).forEach(System.out::print);
		System.out.println();
	}
	
	private int[] primes = {	// primes from https://planetmath.org/goodhashtableprimes
		53,						// we can enumerate all primes that will be used as there is a limit on array size (MAX_VALUE - 5)
		97,
		193,
		389,
		769,
		1543,
		3079,
		6151,
		12289,
		24593,
		49157,
		98317,
		196613,
		393241,
		786433,
		1572869,
		3145739,
		6291469,
		12582917,
		25165843,
		50331653,
		100663319,
		201326611,
		402653189,
		805306457,
		1610612741,
		2147483423
	};
	private int primeIndex = 0;
	private int getNextPrime() {
		if (primeIndex >= primes.length) throw new IndexOutOfBoundsException("Maximum prime array size reached");
		return primes[primeIndex++];
	}

}
